@@ -34,16 +34,27 @@
  |
  |	This notice must be included with this software and any derivatives.
  |
- |	Copyright (c) 1986, 1987, 1988, 1991, 1992, 1993, 1994 Hugh Mahon
+ |	Copyright (c) 1986, 1987, 1988, 1991, 1992, 1993, 1994, 1995 Hugh Mahon
  |	All are rights reserved.
  |
- |	$Header: /home/hugh/sources/old_ae/new_curse.c,v 1.28 1994/11/21 05:58:13 hugh Exp $
+ |	$Header: /home/hugh/sources/old_ae/RCS/new_curse.c,v 1.43 1996/03/21 04:27:06 hugh Exp $
  |
  */
 
+char *copyright_message[] = { "Copyright (c) 1986, 1987, 1988, 1991, 1992, 1993, 1994, 1995 Hugh Mahon",
+				"All rights are reserved."};
+
+char * new_curse_name= "@(#) new_curse.c $Revision: 1.43 $";
+
 #include "new_curse.h"
-#include <signal.h>
 #include <fcntl.h>
+
+#ifdef SYS5
+#include <string.h>
+#else
+#include <strings.h>
+#endif
+
 #ifdef BSD_SELECT
 #include <sys/types.h>
 #include <sys/time.h>
@@ -54,15 +65,28 @@
 
 #endif /* BSD_SELECT */
 
-char *copyright_message[] = { "Copyright (c) 1986, 1987, 1988, 1991, 1992 Hugh Mahon",
-				"All rights are reserved."};
+#ifdef HAS_STDLIB
+#include <stdlib.h>
+#endif
+
+#if defined(__STDC__)
+#include <stdarg.h>
+#else
+#include <varargs.h>
+#endif
+
+#ifdef HAS_UNISTD
+#include <unistd.h>
+#endif
+
+#ifdef HAS_SYS_IOCTL
+#include <sys/ioctl.h>
+#endif
 
-char * new_curse_name= "@(#) new_curse.c $Revision: 1.28 $ pl 4";
 
 WINDOW *curscr;
 static WINDOW *virtual_scr;
 WINDOW *stdscr;
-WINDOW *newwin();
 WINDOW *last_window_refreshed;
 
 #ifdef TIOCGWINSZ
@@ -528,16 +552,45 @@ char *String_table[1024];
 
 int *virtual_lines;
 
+static char nc_scrolling_ability = FALSE;
+
+#ifdef CAP
+
+#if defined(__STDC__) || defined(__cplusplus)
+#define P_(s) s
+#else
+#define P_(s) ()
+#endif /* __STDC__ */
+
+int tc_Get_int P_((int));
+void CAP_PARSE P_((void));
+void Find_term P_((void));
+
+#undef P_
+
+#endif /* CAP */
+
+
+#ifndef __STDC__
+#ifndef HAS_STDLIB
+extern char *fgets();
+extern char *malloc();
+extern char *getenv();
+FILE *fopen();			/* declaration for open function	*/
+#endif /* HAS_STDLIB */
+#endif /* __STDC__ */
+
 #ifdef SIGWINCH
 
 /*
  |	Copy the contents of one window to another.
  */
 
+void 
 copy_window(origin, destination)
 WINDOW *origin, *destination;
 {
-	int counter, row, column;
+	int row, column;
 	struct _line *orig, *dest;
 
 	orig = origin->first_line;
@@ -568,7 +621,9 @@ WINDOW *origin, *destination;
 	destination->SCROLL_CLEAR = origin->SCROLL_CLEAR;
 }
 
-reinitscr() 
+void 
+reinitscr(foo)
+int foo; 
 {
 	WINDOW *local_virt;
 	WINDOW *local_std;
@@ -579,7 +634,7 @@ reinitscr()
 	if (ioctl(0, TIOCGWINSZ, &ws) >= 0)
 	{
 		if (ws.ws_row == LINES && ws.ws_col == COLS) 
-			return(0);
+			return;
 		if (ws.ws_row > 0) 
 			LINES = ws.ws_row;
 		if (ws.ws_col > 0) 
@@ -604,12 +659,15 @@ reinitscr()
 }
 #endif /* SIGWINCH */
 
+void 
 initscr()		/* initialize terminal for operations	*/
 {
-	char *pointer;
 	int value;
 	char *lines_string;
 	char *columns_string;
+#ifdef CAP
+	char *pointer;
+#endif /* CAP */
 
 #ifdef DIAG
 printf("starting initscr \n");fflush(stdout);
@@ -848,9 +906,19 @@ printf("starting initscr \n");fflush(stdout);
 	 */
 	signal(SIGWINCH, reinitscr);
 #endif /* SIGWINCH */
+
+	/*
+	 |	check if scrolling is available
+	 */
+
+	nc_scrolling_ability = ((String_table[al__] != NULL) && 
+				(String_table[dl__])) || ((String_table[cs__]) 
+				&& (String_table[sr__]));
+
 }
 
 #ifndef CAP
+int 
 Get_int()		/* get a two-byte integer from the terminfo file */
 {
 	int High_byte;
@@ -871,6 +939,7 @@ Get_int()		/* get a two-byte integer from the terminfo file */
 		return(Low_byte + (High_byte * 256));
 }
 
+int 
 INFO_PARSE()		/* parse off the data in the terminfo data file	*/
 {
 	int offset;
@@ -883,7 +952,7 @@ INFO_PARSE()		/* parse off the data in the terminfo data file	*/
 	int string_table_len = 0;
 	char *temp_ptr;
 
-	TERM_data_ptr = Data_Line = calloc(1, (10240 * (sizeof(char))));
+	TERM_data_ptr = Data_Line = malloc((10240 * (sizeof(char))));
 	Data_Line_len = read(Fildes, Data_Line, 10240);
 	if ((Data_Line_len >= 10240) || (Data_Line_len < 0))
 		return(0);
@@ -911,7 +980,7 @@ INFO_PARSE()		/* parse off the data in the terminfo data file	*/
 	Num_ints = Get_int();
 	Num_strings = Get_int();
 	string_table_len = Get_int();
-	Strings = calloc(1, string_table_len);
+	Strings = malloc(string_table_len);
 	while (Num_names > 0)
 	{
 		TERM_data_ptr++;
@@ -948,11 +1017,12 @@ INFO_PARSE()		/* parse off the data in the terminfo data file	*/
 		counter++;
 	}
 	close(Fildes);
-	cfree(Data_Line);
+	free(Data_Line);
 	return(TRUE);
 }
 #endif		/* ifndef CAP	*/
 
+int 
 AtoI()		/* convert ascii text to integers	*/
 {
 	int Temp;
@@ -966,6 +1036,7 @@ AtoI()		/* convert ascii text to integers	*/
 	return(Temp);
 }
 
+void 
 Key_Get()		/* create linked list with all key sequences obtained from terminal database	*/
 {
 	int Counter;
@@ -1006,6 +1077,34 @@ Key_Get()		/* create linked list with all key sequences obtained from terminal d
 			Klen = strlen(Spoint->element->string);
 			if (Klen > Max_Key_len)
 				Max_Key_len = Klen;
+			/*
+			 |  Some terminal types accept keystrokes of the form
+			 |  \E[A and \EOA, substituting '[' for 'O'.  Make a 
+			 |  duplicate of such key strings (since the 
+			 |  database will only have one version) so new_curse 
+			 |  can understand both.
+			 */
+			if ((Spoint->element->length > 1) && 
+			    ((String_table[key_def][1] == '[') || 
+			     (String_table[key_def][1] == 'O')))
+			{
+				Spoint->next = (struct KEY_STACK *) malloc(sizeof(struct KEY_STACK));
+				Spoint = Spoint->next;
+				Spoint->next = NULL;
+				Spoint->element = (struct KEYS *) malloc(sizeof(struct KEYS));
+				Spoint->element->length = strlen(String_table[key_def]);
+				Spoint->element->string = malloc(Spoint->element->length + 1);
+				strcpy(Spoint->element->string, String_table[key_def]);
+				Spoint->element->value = Key_vals[Counter];
+				Klen = strlen(Spoint->element->string);
+				if (Klen > Max_Key_len)
+					Max_Key_len = Klen;
+			
+				if (String_table[key_def][1] == '[')
+					Spoint->element->string[1] = 'O';
+				else
+					Spoint->element->string[1] = '[';
+			}
 		}
 		key_def++;
 		Counter++;
@@ -1013,7 +1112,8 @@ Key_Get()		/* create linked list with all key sequences obtained from terminal d
 }
 
 #ifdef CAP
-char *String_Get(param)		/* read the string */
+char *
+String_Get(param)		/* read the string */
 char *param;
 {
 	char *String;
@@ -1026,7 +1126,7 @@ char *param;
 			TERM_data_ptr++;
 		Temp = ++TERM_data_ptr;
 		Counter = 1;
-		while ((*Temp != ':') && (*Temp != NULL))
+		while ((*Temp != ':') && (*Temp != (char)NULL))
 		{
 			Counter++;
 			Temp++;
@@ -1034,7 +1134,7 @@ char *param;
 		if (Counter == 1)	/* no data */
 			return(NULL);
 		String = Temp = malloc(Counter);
-		while ((*TERM_data_ptr != ':') && (*TERM_data_ptr != NULL))
+		while ((*TERM_data_ptr != ':') && (*TERM_data_ptr != (char)NULL))
 		{
 			if (*TERM_data_ptr == '\\')
 			{
@@ -1061,7 +1161,7 @@ char *param;
 					while ((*TERM_data_ptr >= '0') && (*TERM_data_ptr <= '9'))
 					{
 						Counter = (8 * Counter) + (*TERM_data_ptr - '0');
-						*TERM_data_ptr++;
+						TERM_data_ptr++;  /* ? */
 					}
 					*Temp = Counter;
 					TERM_data_ptr--;
@@ -1082,26 +1182,26 @@ char *param;
 			else
 				*Temp++ = *TERM_data_ptr++;
 		}
-		*Temp = NULL;
+		*Temp = (char)NULL;
 		param = String;
 	}
 	else
 	{
-		while ((*TERM_data_ptr != NULL) && (*TERM_data_ptr != ':'))
+		while ((*TERM_data_ptr != (char)NULL) && (*TERM_data_ptr != ':'))
 			TERM_data_ptr++;
 	}
 	return(param);
 }
 
-Get_int(param)		/* read the integer				*/
+int 
+tc_Get_int(param)		/* read the integer			*/
 int param;
 {
-	char *Temp;
 	int Itemp;
 
 	if (param == 0)
 	{
-		while ((*TERM_data_ptr != NULL) && (*TERM_data_ptr != '#'))
+		while ((*TERM_data_ptr != (char)NULL) && (*TERM_data_ptr != '#'))
 			TERM_data_ptr++;
 		TERM_data_ptr++;
 		Itemp = AtoI();
@@ -1115,6 +1215,7 @@ int param;
 	return(param);
 }
 
+void 
 Find_term()		/* find terminal description in termcap file	*/
 {
 	char *Name;
@@ -1122,25 +1223,25 @@ Find_term()		/* find terminal description in termcap file	*/
 
 	Ftemp = Name = malloc(strlen(TERMINAL_TYPE + 1) + 1);
 	strcpy(Name, TERMINAL_TYPE);
-	while (*Ftemp != NULL)
+	while (*Ftemp != (char)NULL)
 		Ftemp++;
 	*Ftemp++ = '|';
-	*Ftemp = NULL;
+	*Ftemp = (char)NULL;
 	CFOUND = FALSE;
 	Data_Line_len = strlen(TERMINAL_TYPE) + 1;
 	while ((!CFOUND) && ((TERM_data_ptr=fgets(Data_Line, 512, TFP)) != NULL))
 	{
 		if ((*TERM_data_ptr != ' ') && (*TERM_data_ptr != '\t') && (*TERM_data_ptr != '#'))
 		{
-			while ((!CFOUND) && (*TERM_data_ptr != NULL))
+			while ((!CFOUND) && (*TERM_data_ptr != (char)NULL))
 			{
 				CFOUND = !strncmp(TERM_data_ptr, Name, Data_Line_len);
-				while ((*TERM_data_ptr != NULL) && (*TERM_data_ptr != '|') && (*TERM_data_ptr != '#') && (*TERM_data_ptr != ':'))
+				while ((*TERM_data_ptr != (char)NULL) && (*TERM_data_ptr != '|') && (*TERM_data_ptr != '#') && (*TERM_data_ptr != ':'))
 					TERM_data_ptr++;
 				if (*TERM_data_ptr == '|')
 					TERM_data_ptr++;
 				else if (!CFOUND)
-					*TERM_data_ptr = NULL;
+					*TERM_data_ptr = (char)NULL;
 			}
 		}
 	}
@@ -1151,15 +1252,15 @@ Find_term()		/* find terminal description in termcap file	*/
 	}
 }
 
+void 
 CAP_PARSE()		/* parse off the data in the termcap data file	*/
 {
 	int offset;
-	int Cip = 0;
 	int found;
 
 	do
 	{
-		while (*TERM_data_ptr != NULL)
+		while (*TERM_data_ptr != (char)NULL)
 		{
 			for (found = FALSE, offset = 0; (!found) && (offset < 26); offset++)
 			{
@@ -1176,7 +1277,7 @@ CAP_PARSE()		/* parse off the data in the termcap data file	*/
 					if (!strncmp(TERM_data_ptr, Number_names[offset], 3))
 					{
 						found = TRUE;
-						Numbers[offset] = Get_int(Numbers[offset]);
+						Numbers[offset] = tc_Get_int(Numbers[offset]);
 					}
 				}
 			}
@@ -1194,7 +1295,7 @@ CAP_PARSE()		/* parse off the data in the termcap data file	*/
 
 			if (!strncmp(TERM_data_ptr, "tc=", 3))
 				tc_ = String_Get(NULL);
-			while ((*TERM_data_ptr != ':') && (*TERM_data_ptr != NULL))
+			while ((*TERM_data_ptr != ':') && (*TERM_data_ptr != (char)NULL))
 				TERM_data_ptr++;
 			if (*TERM_data_ptr == ':')
 				TERM_data_ptr++;
@@ -1260,12 +1361,17 @@ int start_l, start_c;	/* starting line and column to be inwindow	*/
 	Ntemp->Attrib = FALSE;
 	Ntemp->first_line = temp_screen = Screenalloc(cols);
 	Ntemp->first_line->number = 0;
+	Ntemp->line_array = (struct _line **) malloc(LINES * sizeof(struct _line *));
+	
+	Ntemp->line_array[0] = Ntemp->first_line;
+
 	for (i = 1; i < lines; i++)
 	{
 		temp_screen->next_screen = Screenalloc(cols);
 		temp_screen->next_screen->number = i;
 		temp_screen->next_screen->prev_screen = temp_screen;
 		temp_screen = temp_screen->next_screen;
+		Ntemp->line_array[i] = temp_screen;
 	}
 	Ntemp->first_line->prev_screen = NULL;
 	temp_screen->next_screen = NULL;
@@ -1273,7 +1379,8 @@ int start_l, start_c;	/* starting line and column to be inwindow	*/
 }
 
 #ifdef CAP
-Cap_Out(string, p_list, place)		/* interpret the output string if necessary */
+void 
+Cap_Out(string, p_list, place)	/* interpret the output string if necessary */
 char *string;
 int p_list[];			/* stack of values	*/
 int place;			/* place keeper of top of stack	*/
@@ -1282,16 +1389,10 @@ int place;			/* place keeper of top of stack	*/
 	int delay;
 	int p1, p2, temp;
 	float chars;
-	int old_arg, value;
 
 	if (string == NULL)
 		return;
 
-/*#ifndef SYS5
-	old_arg = fcntl(0, F_GETFL, 0);
-	value = fcntl(0, F_SETFL, old_arg & ~FNDELAY);
-#endif*/
-
 	if (p_list != NULL)
 	{
 		p1 = p_list[--place];
@@ -1307,7 +1408,7 @@ int place;			/* place keeper of top of stack	*/
 		if (*Otemp == '*')
 			Otemp++;
 	}
-	while (*Otemp != NULL)
+	while (*Otemp != (char)NULL)
 	{
 		if (*Otemp == '%')
 		{
@@ -1385,10 +1486,6 @@ int place;			/* place keeper of top of stack	*/
 			putchar(*String_table[pc__]);
 	}
 	fflush(stdout);
-/*#ifndef SYS5
-	old_arg = fcntl(0, F_GETFL, 0);
-	value = fcntl(0, F_SETFL, old_arg | FNDELAY);
-#endif*/
 }
 
 #else
@@ -1398,7 +1495,8 @@ int place;			/* place keeper of top of stack	*/
 	int p[10];
 	int variable[27];
 
-Operation(Temp_Stack, place)		/* handle conditional operations	*/
+int 
+Operation(Temp_Stack, place)	/* handle conditional operations	*/
 int Temp_Stack[];
 int place;
 {
@@ -1582,6 +1680,7 @@ int place;
 	return(place);
 }
 
+void 
 Info_Out(string, p_list, place)	/* interpret the output string if necessary */
 char *string;
 int p_list[];
@@ -1590,7 +1689,6 @@ int place;
 	char *tchar;
 	int delay;
 	int temp;
-	int value;
 	int Cond_FLAG;
 	int EVAL;
 	int Cond_Stack[128];
@@ -1777,22 +1875,7 @@ int place;
 }
 #endif
 
-atoi(string)		/* convert ascii text to integers	*/
-char *string;
-{
-	int Temp;
-	char *astring;
-
-	astring = string;
-	Temp = 0;
-	while ((*astring >= '0') && (*astring <= '9'))
-	{
-		Temp = (Temp * 10) + (*astring - '0');
-		astring++;
-	}
-	return(Temp);
-}
-
+void 
 wmove(window, row, column)	/* move cursor to indicated position in window */
 WINDOW *window;
 int row, column;
@@ -1804,27 +1887,29 @@ int row, column;
 	}
 }
 
+void 
 clear_line(line, column, cols)
 struct _line *line;
 int column;
+int cols;
 {
 	int j;
 
 	if (column > line->last_char)
-		line->row[line->last_char] = ' ';
+	{
+		for (j = line->last_char; j < column; j++)
+		{
+			line->row[j] = ' ';
+			line->attributes[j] = (char) NULL;
+		}
+	}
 	line->last_char = column;
 	line->row[column] = (char) NULL;
 	line->attributes[column] = (char) NULL;
 	line->changed = TRUE;
-/*	if (column == 0)
-		line->scroll = FALSE;*/
-	for (j = column + 1; j < cols; j++)
-	{
-		line->row[j] = ' ';
-		line->attributes[j] = (char) NULL;
-	}
 }
 
+void 
 werase(window)			/* clear the specified window		*/
 WINDOW *window;
 {
@@ -1837,6 +1922,7 @@ WINDOW *window;
 		clear_line(tmp, 0, window->Num_cols);
 }
 
+void 
 wclrtoeol(window)	/* erase from current cursor position to end of line */
 WINDOW *window;
 {
@@ -1851,18 +1937,15 @@ WINDOW *window;
 	clear_line(tmp, column, window->Num_cols);
 }
 
+void 
 wrefresh(window)		/* flush all previous output		*/
 WINDOW *window;
 {
-	int old_arg, value;
-/*#ifndef SYS5
-	old_arg = fcntl(0, F_GETFL, 0);
-	value = fcntl(0, F_SETFL, old_arg & ~FNDELAY);
-#endif   to be uncommented if setting to nodelay on input only is unsuitable */
 	wnoutrefresh(window);
 #ifdef DIAG
 {
 	struct _line *temp;
+	int value;
 	fprintf(stderr, "columns=%d, lines=%d, SC=%d, SR=%d\n",window->Num_cols, window->Num_lines, window->SC, window->SR);
 	for (value = 0, temp = window->first_line; value < window->Num_lines; value++, temp = temp->next_screen)
 	{
@@ -1891,11 +1974,9 @@ WINDOW *window;
 	virtual_scr->SCROLL_CLEAR = FALSE;
 	virtual_scr->scroll_down = virtual_scr->scroll_up = 0;
 	fflush(stdout);
-/*#ifndef SYS5
-	value = fcntl(0, F_SETFL, old_arg | FNDELAY);
-#endif*/
 }
 
+void 
 touchwin(window)
 WINDOW *window;
 {
@@ -1910,6 +1991,7 @@ WINDOW *window;
 	window->SCROLL_CLEAR = TRUE;
 }
 
+void 
 wnoutrefresh(window)
 WINDOW *window;
 {
@@ -1942,12 +2024,21 @@ WINDOW *window;
 		{
 			for (user_col = 0, virt_col = window->SC; 
 				(virt_col < virtual_scr->Num_cols) 
-				  && (user_col < window->Num_cols); 
+				  && (user_col < user_line->last_char); 
 				  	virt_col++, user_col++)
 			{
 				virtual_line->row[virt_col] = user_line->row[user_col];
 				virtual_line->attributes[virt_col] = user_line->attributes[user_col];
 			}
+			for (user_col = user_line->last_char, 
+			     virt_col = window->SC + user_line->last_char; 
+				(virt_col < virtual_scr->Num_cols) 
+				  && (user_col < window->Num_cols); 
+				  	virt_col++, user_col++)
+			{
+				virtual_line->row[virt_col] = ' ';
+				virtual_line->attributes[virt_col] = (char) NULL;
+			}
 		}
 		if (virtual_scr->Num_cols != window->Num_cols)
 		{
@@ -1959,11 +2050,6 @@ WINDOW *window;
 					min(virtual_scr->Num_cols, 
 					  (user_line->last_char + window->SC));
 			}
-			else if (virtual_line->last_char > (user_line->last_char + window->SC))
-			{
-				virtual_line->row[min(virtual_scr->Num_cols, 
-					  (user_line->last_char + window->SC))] = ' ';
-			}
 		}
 		else
 			virtual_line->last_char = user_line->last_char;
@@ -1977,18 +2063,12 @@ WINDOW *window;
 	last_window_refreshed = window;
 }
 
+void 
 flushinp()			/* flush input				*/
 {
-	int value;
-/*
-#ifdef SYS5
-	value = ioctl(0, TCFLSH, 0);
-#else
-	value = ioctl(0, TIOCFLUSH, 0);
-#endif
-*/
 }
 
+void 
 ungetch(c)			/* push a character back on input	*/
 int c;
 {
@@ -2029,12 +2109,15 @@ timed_getchar()
 }
 #endif
 
+int 
 wgetch(window)			/* get character from specified window	*/
 WINDOW *window;
 {
-	int old_arg;
 	int in_value;
 	char temp;
+#ifndef SYS5
+	int old_arg;
+#endif /* SYS5 */
 
 #ifdef BSD_SELECT
 	if (Noblock)
@@ -2043,7 +2126,8 @@ WINDOW *window;
 		in_value = ((bufp > 0) ? in_buff[--bufp] : read(0, &temp, 1)? temp : -1);
 #else /* BSD_SELECT */
 #ifdef SYS5
-	in_value = ((bufp > 0) ? in_buff[--bufp] : read(0, &temp, 1)? temp : -1);
+	in_value = ((bufp > 0) ? in_buff[--bufp] : 
+					(read(0, &temp, 1)> 0) ? temp : -1);
 #else /* SYS5 */
 	if (Noblock)
 	{
@@ -2080,14 +2164,15 @@ WINDOW *window;
 }
 
 #ifndef BSD_SELECT
-Clear()		/* notify that time out has occurred	*/
+void 
+Clear(arg)		/* notify that time out has occurred	*/
+int arg;
 {
 	Time_Out = TRUE;
 #ifdef DEBUG
 fprintf(stderr, "inside Clear()\n");
 fflush(stderr);
 #endif /* DEBUG */
-
 }
 #endif /* BSD_SELECT */
 
@@ -2099,15 +2184,16 @@ int first_char;				/* first character of sequence	*/
 	int Count;
 	char string[128];
 	char *Gtemp;
-	int value;
 	int Found;
-	int old_arg;
 #ifdef SYS5
 	struct termio Gterminal;
 #else
 	struct sgttyb Gterminal;
 #endif
 	struct KEY_STACK *St_point;
+#if (!defined( BSD_SELECT)) || (!defined(SYS5))
+	int value;
+#endif /* BSD_SELECT */
 
 	Count = 0;
 	Gtemp = string;
@@ -2120,19 +2206,9 @@ int first_char;				/* first character of sequence	*/
 #endif /* BSD_SELECT */
 	Noblock = TRUE;
 #ifdef SYS5
-/*	value = ioctl(0, TCGETA, &Gterminal);
-	Gterminal.c_cc[VMIN] = 0;	/* minimum of zero characters	*/
 	Gterminal.c_cc[VTIME] = 0;		/* timeout value	*/
 	Gterminal.c_lflag &= ~ICANON;	/* disable canonical operation	*/
 	Gterminal.c_lflag &= ~ECHO;		/* disable echo		*/
-/*	value = ioctl(0, TCSETA, &Gterminal);	/* set characteristics	*/
-#else
-/*	value = ioctl(0, TIOCGETP, &Gterminal);
-	Gterminal.sg_flags |= RAW ;
-	Gterminal.sg_flags &= ~ECHO;
-	value = ioctl(0, TIOCSETP, &Gterminal);	/* set characteristics	*/
-/*	old_arg = fcntl(0, F_GETFL, 0);
-	value = fcntl(0, F_SETFL, old_arg | FNDELAY);*/
 #endif
 	Count = 1;
 	Found = FALSE;
@@ -2189,19 +2265,18 @@ fprintf(stderr, "ungetting character %d\n", string[Count]);fflush(stdout);
 	}
 }
 
+void 
 waddch(window, c)	/* output the character in the specified window	*/
 WINDOW *window;
 int c;
 {
-	int i, j;
-	int row, column;
+	int column, j;
 	int shift;	/* number of spaces to shift if a tab		*/
 	struct _line *tmpline;
 
 #ifdef DIAG
 /*printf("starting waddch \n");fflush(stdout);*/
 #endif
-	row = window->LY;
 	column = window->LX;
 	if (c == '\t')
 	{
@@ -2216,27 +2291,25 @@ int c;
 			waddch(window, ' ');
 		}
 	}
-	else if ((column < window->Num_cols) && (row < window->Num_lines))
+	else if ((column < window->Num_cols) && (window->LY < window->Num_lines))
 	{
 		if ((c == '~') && (Booleans[hz__]))
 			c = '@';
 
 		if (( c != '\b') && (c != '\n') && (c != '\r'))
 		{
-			row = 0;
-			tmpline = window->first_line; 
-			while (row < window->LY)
-			{
-				row++;
-				tmpline = tmpline->next_screen;
-			}
+			tmpline = window->line_array[window->LY];
 			tmpline->row[column] = c;
 			tmpline->attributes[column] = window->Attrib;
 			tmpline->changed = TRUE;
 			if (column >= tmpline->last_char)
 			{
 				if (column > tmpline->last_char)
-					tmpline->row[tmpline->last_char] = ' ';
+					for (j = tmpline->last_char; j < column; j++)
+					{
+						tmpline->row[j] = ' ';
+						tmpline->attributes[j] = (char) NULL;
+					}
 				tmpline->row[column + 1] = (char) NULL;
 				tmpline->attributes[column + 1] = (char) NULL;
 				tmpline->last_char = column + 1;
@@ -2270,6 +2343,7 @@ int c;
 	window->SCROLL_CLEAR = CHANGE;
 }
 
+void 
 winsertln(window)	/* insert a blank line into the specified window */
 WINDOW *window;
 {
@@ -2309,8 +2383,16 @@ WINDOW *window;
 	}
 	if (window->LY == 0)
 		window->first_line = tmp1;
+
+	for (row = 0, tmp1 = window->first_line; 
+		row < window->Num_lines; row++)
+	{
+		window->line_array[row] = tmp1;
+		tmp1 = tmp1->next_screen;
+	}
 }
 
+void 
 wdeleteln(window)	/* delete a line in the specified window */
 WINDOW *window;
 {
@@ -2351,6 +2433,12 @@ WINDOW *window;
 		else
 			tmp = tmpline;
 		tmp->next_screen = NULL;
+
+		for (row = 0, tmp = window->first_line; row < window->Num_lines; row++)
+		{
+			window->line_array[row] = tmp;
+			tmp = tmp->next_screen;
+		}
 	}
 	else
 	{
@@ -2358,11 +2446,11 @@ WINDOW *window;
 	}
 }
 
+void 
 wclrtobot(window)	/* delete from current position to end of the window */
 WINDOW *window;
 {
 	int row, column;
-	int Trow;
 	struct _line *tmp;
 
 	window->SCROLL_CLEAR |= CLEAR;
@@ -2379,6 +2467,7 @@ WINDOW *window;
 	wmove(window, row, column);
 }
 
+void 
 wstandout(window)	/* begin standout mode in window	*/
 WINDOW *window;
 {
@@ -2386,12 +2475,14 @@ WINDOW *window;
 		window->Attrib |= A_STANDOUT;
 }
 
+void 
 wstandend(window)	/* end standout mode in window	*/
 WINDOW *window;
 {
 	window->Attrib &= ~A_STANDOUT;
 }
 
+void 
 waddstr(window, string)	/* write 'string' in window	*/
 WINDOW *window;
 char *string;
@@ -2402,6 +2493,7 @@ char *string;
 		waddch(window, *wstring);
 }
 
+void 
 clearok(window, flag)	/* erase screen and redraw at next refresh	*/
 WINDOW *window;
 int flag;
@@ -2409,6 +2501,7 @@ int flag;
 	Repaint_screen = TRUE;
 }
 
+void 
 echo()			/* turn on echoing				*/
 {
 	int value;
@@ -2422,6 +2515,7 @@ echo()			/* turn on echoing				*/
 #endif
 }
 
+void 
 noecho()		/* turn off echoing				*/
 {
 	int value;
@@ -2435,10 +2529,10 @@ noecho()		/* turn off echoing				*/
 #endif
 }
 
+void 
 raw()			/* set to read characters immediately		*/
 {
 	int value;
-	int old_arg;
 
 #ifdef SYS5
 	Intr = Terminal.c_cc[VINTR];	/* get the interrupt character	*/
@@ -2460,15 +2554,13 @@ raw()			/* set to read characters immediately		*/
 #else
 	Terminal.sg_flags |= RAW;	/* enable raw mode		*/
 	value = ioctl(0, TIOCSETP, &Terminal);	/* set characteristics	*/
-/*	old_arg = fcntl(0, F_GETFL, 0);
-	value = fcntl(0, F_SETFL, old_arg | FNDELAY);*/
 #endif
 }
 
+void 
 noraw()			/* set to normal character read mode		*/
 {
 	int value;
-	int old_arg;
 
 #ifdef SYS5
 	Terminal.c_lflag |= ICANON;	/* enable canonical operation	*/
@@ -2485,6 +2577,7 @@ noraw()			/* set to normal character read mode		*/
 #endif
 }
 
+void 
 nl()
 {
 	int value;
@@ -2495,6 +2588,7 @@ nl()
 #endif
 }
 
+void 
 nonl()
 {
 	int value;
@@ -2506,30 +2600,36 @@ nonl()
 #endif
 }
 
+void 
 saveterm()
 {
 }
 
+void 
 fixterm()
 {
 }
 
+void 
 resetterm()
 {
 }
 
+void 
 nodelay(window, flag)
 WINDOW *window;
 int flag;
 {
 }
 
+void 
 idlok(window, flag)
 WINDOW *window;
 int flag;
 {
 }
 
+void 
 keypad(window, flag)
 WINDOW *window;
 int flag;
@@ -2540,6 +2640,7 @@ int flag;
 		String_Out(String_table[ke__], NULL, 0);
 }
 
+void 
 savetty()		/* save current tty stats			*/
 {
 	int value;
@@ -2551,6 +2652,7 @@ savetty()		/* save current tty stats			*/
 #endif
 }
 
+void 
 resetty()		/* restore previous tty stats			*/
 {
 	int value;
@@ -2562,6 +2664,7 @@ resetty()		/* restore previous tty stats			*/
 #endif
 }
 
+void 
 endwin()		/* end windows					*/
 {
 	keypad(stdscr, FALSE);
@@ -2579,6 +2682,7 @@ endwin()		/* end windows					*/
 #endif
 }
 
+void 
 delwin(window)		/* delete the window structure			*/
 WINDOW *window;
 {
@@ -2602,20 +2706,35 @@ WINDOW *window;
 	}
 }
 
+#ifndef __STDC__
+void 
 wprintw(va_alist)
 va_dcl
+#else /* __STDC__ */
+void 
+wprintw(WINDOW *window, const char *format, ...)
+#endif /* __STDC__ */
 {
+#ifndef __STDC__
 	WINDOW *window;
-	va_list ap;
 	char *format;
+	va_list ap;
+#else
+	va_list ap;
+#endif
 	int value;
 	char *fpoint;
 	char *wtemp;
 
+#ifndef __STDC__
 	va_start(ap);
 	window = va_arg(ap, WINDOW *);
 	format = va_arg(ap, char *);
-	fpoint = format;
+#else /* __STDC__ */
+	va_start(ap, format);
+#endif /* __STDC__ */
+
+	fpoint = (char *) format;
 	while (*fpoint != (char) NULL)
 	{
 		if (*fpoint == '%')
@@ -2658,8 +2777,12 @@ va_dcl
 		else
 			waddch(window, *fpoint++);
 	}
+#ifdef __STDC__
+	va_end(ap);
+#endif /* __STDC__ */
 }
 
+void 
 iout(window, value)	/* output characters		*/
 WINDOW *window;
 int value;
@@ -2671,45 +2794,40 @@ int value;
 	waddch(window, ((value % 10) + '0'));
 }
 
+int 
 Comp_line(line1, line2)		/* compare lines	*/
 struct _line *line1;
 struct _line *line2;
 {
-	int count1, count2;
+	int count1;
 	int i;
 	char *att1, *att2;
 	char *c1, *c2;
 
+	if (line1->last_char != line2->last_char)
+		return(2);
+
 	c1 = line1->row;
 	c2 = line2->row;
 	att1 = line1->attributes;
 	att2 = line2->attributes;
-	count2 = strlen(c1) + 1;
-	count1 = strlen(c2) + 1;
-	if (count1 > count2)
-	{
-		i = count2;
-		count2 = count1;
-		count1 = i;
-	}
-	if (count2 > (count1 + count1))
-		return(2);
 	i = 0;
 	while ((c1[i] != (char) NULL) && (c2[i] != (char) NULL) && (c1[i] == c2[i]) && (att1[i] == att2[i]))
 		i++;
 	count1 = i + 1;
-	if ((count1 == 1) && (count2 == 1))
+	if ((count1 == 1) && (c1[i] == (char) NULL) && (c2[i] == (char) NULL))
 		count1 = 0;			/* both lines blank	*/
-	else if (count2 == count1)
+	else if ((c1[i] == (char) NULL) && (c2[i] == (char) NULL))
 		count1 = -1;			/* equal		*/
 	else
-		count1 = count2 / count1;	/* lines unequal	*/
+		count1 = 1;			/* lines unequal	*/
 	return(count1);
 }
 
 struct _line *
-Insert_line(row, window)	/* insert line into screen */
+Insert_line(row, end_row, window)	/* insert line into screen */
 int row;
+int end_row;
 WINDOW *window;
 {
 	int i;
@@ -2718,20 +2836,22 @@ WINDOW *window;
 
 	for (i = 0, tmp = curscr->first_line; i < window->SR; i++)
 		tmp = tmp->next_screen;
+	if ((end_row + window->SR) == 0)
+		curscr->first_line = curscr->first_line->next_screen;
 	top_of_win = tmp;
 	/*
 	 |	find bottom line to delete
 	 */
-	for (i = 1, tmp = top_of_win; (tmp->next_screen != NULL) && (i < window->Num_lines); i++)
+	for (i = 0, tmp = top_of_win; (tmp->next_screen != NULL) && (i < end_row); i++)
 		tmp = tmp->next_screen;
 	if (tmp->prev_screen != NULL)
 		tmp->prev_screen->next_screen = tmp->next_screen;
 	if (tmp->next_screen != NULL)
 		tmp->next_screen->prev_screen = tmp->prev_screen;
+	tmp1 = tmp;
 	/*
 	 |	clear deleted line
 	 */
-	tmp1 = tmp;
 	clear_line(tmp, 0, window->Num_cols);
 	tmp1->number = -1;
 	for (i = 0, tmp = curscr->first_line; (tmp->next_screen != NULL) && (i < window->SR); i++)
@@ -2748,18 +2868,28 @@ WINDOW *window;
 		curscr->first_line = tmp1;
 	if (tmp1->next_screen != NULL)
 		tmp1 = tmp1->next_screen;
-/*	Position(window, (window->Num_lines -1), 0);
-	String_Out(String_table[dl__], NULL, 0);*/
-	Position(window, row, 0);
-	String_Out(String_table[al__], NULL, 0);
+
+	if ((!String_table[cs__]) && (end_row < window->Num_lines))
+	{
+		Position(window, (window->SR + end_row), 0);
+		String_Out(String_table[dl__], NULL, 0);
+	}
+	Position(window, (window->SR + row), 0);
+	if (String_table[al__] != NULL)
+		String_Out(String_table[al__], NULL, 0);
+	else
+		String_Out(String_table[sr__], NULL, 0);
+
 	for (i = 0, top_of_win = curscr->first_line; (top_of_win->next_screen != NULL) && (i < window->SR); i++)
 		top_of_win = top_of_win->next_screen;
 	return(tmp1);
 }
 
+
 struct _line *
-Delete_line(row, window)	/* delete a line on screen */
+Delete_line(row, end_row, window)	/* delete a line on screen */
 int row;
+int end_row;
 WINDOW *window;
 {
 	int i;
@@ -2799,17 +2929,27 @@ WINDOW *window;
 	for (i = 0, tmp = curscr->first_line; (tmp->next_screen != NULL) && (i < window->SR); i++)
 		tmp = tmp->next_screen;
 	top_of_win = tmp;
-	for (i = 1, tmp = top_of_win; (i < window->Num_lines) && (tmp->next_screen != NULL); i++)
+	for (i = 0, tmp = top_of_win; (i < end_row) && (tmp->next_screen != NULL); i++)
 		tmp = tmp->next_screen;
-	tmp1->next_screen = tmp->next_screen;
-	tmp1->prev_screen = tmp;
-	if (tmp1->next_screen != NULL)
-		tmp1->next_screen->prev_screen = tmp1;
-	tmp->next_screen = tmp1;
-	Position(window, row, 0);
+	tmp1->next_screen = tmp;
+	tmp1->prev_screen = tmp->prev_screen;
+	if (tmp1->prev_screen != NULL)
+		tmp1->prev_screen->next_screen = tmp1;
+	tmp->prev_screen = tmp1;
+
+	Position(window, (window->SR + row), 0);
 	String_Out(String_table[dl__], NULL, 0);
-/*	Position(window, (window->Num_lines -1), 0);
-	String_Out(String_table[al__], NULL, 0);*/
+	if ((!String_table[cs__]) && (end_row < window->Num_lines))
+	{
+		Position(window, (window->SR + end_row), 0);
+		String_Out(String_table[al__], NULL, 0);
+	}
+	else if ((String_table[cs__] != NULL) && (String_table[dl__] == NULL))
+	{
+		Position(window, (window->SR + end_row), 0);
+		putchar('\n');
+	}
+
 	if (row == (window->Num_lines-1))
 		tmp2 = tmp1;
 	if ((row + window->SR) == 0)
@@ -2817,6 +2957,7 @@ WINDOW *window;
 	return(tmp2);
 }
 
+void 
 CLEAR_TO_EOL(window, row, column)
 WINDOW *window;
 int row, column;
@@ -2852,6 +2993,7 @@ int row, column;
 	}
 }
 
+int 
 check_delete(window, line, offset, pointer_new, pointer_old)
 WINDOW *window;
 int line, offset;
@@ -2860,7 +3002,6 @@ struct _line *pointer_new, *pointer_old;
 	int end_old;
 	int end_new;
 	int k;
-	int i;
 	int changed;
 	char *old_lin;
 	char *new_lin;
@@ -2899,16 +3040,15 @@ struct _line *pointer_new, *pointer_old;
 	return(changed);
 }
 
+int 
 check_insert(window, line, offset, pointer_new, pointer_old)
 WINDOW *window;
 int line, offset;
 struct _line *pointer_new, *pointer_old;
 {
 	int changed;
-	int count1, count2;
-	int begin_old, begin_new;
 	int end_old, end_new;
-	int i, j, k;
+	int k;
 	int same = FALSE;
 	int old_off;
 	int insert;
@@ -2966,6 +3106,7 @@ struct _line *pointer_new, *pointer_old;
 	return(changed);
 }
 
+void 
 doupdate()
 {
 	WINDOW *window;
@@ -2979,7 +3120,8 @@ doupdate()
 	int first_time;
 	int first_same;
 	int last_same;
-
+	int list[10];
+	int bottom;
 
 	struct _line *curr;
 	struct _line *virt;
@@ -2987,6 +3129,8 @@ doupdate()
 
 	struct _line *new;
 
+	struct _line *old1, *new1;
+
 	char *cur_lin;
 	char *vrt_lin;
 	char *cur_att;
@@ -3043,8 +3187,8 @@ doupdate()
 		virtual_lines[from_top] = TRUE;
 		if ((similar = Comp_line(curr, virt)) > 0)
 		{
-			diff = TRUE;
 			virtual_lines[from_top] = FALSE;
+			diff = TRUE;
 		}
 		curr = curr->next_screen;
 		virt = virt->next_screen;
@@ -3061,54 +3205,88 @@ doupdate()
 	{
 		last_same = -1;
 		changed = FALSE;
-		while ((from_top < window->Num_lines) && (String_table[al__]) && (String_table[dl__]))	/* check entire lines for diffs	*/
+		for (first_same = window->Num_lines; 
+		    (first_same > from_top) && (virtual_lines[first_same - 1]);
+		     first_same--)
+			;
+		for (last_same = 0;
+		    (last_same < window->Num_lines) && (virtual_lines[last_same]== FALSE);
+		     last_same++)
+			;
+		while ((from_top < first_same) && nc_scrolling_ability)
+					/* check entire lines for diffs	*/
 		{
 
+			if (from_top >= last_same)
+			{
+				for (last_same = from_top; 
+				     (last_same < window->Num_lines) && 
+				     (virtual_lines[last_same] == FALSE);
+				      last_same++)
+					;
+			}
 			if (!virtual_lines[from_top])
 			{
-				for (first_same = window->Num_lines; (first_same > from_top) && (virtual_lines[first_same - 1]); first_same--)
-					;
-				count1 = first_same - 1;
 				diff = TRUE;
 				/*
 				 |	check for lines deleted (scroll up)
 				 */
 				for (tmp_ft = from_top+1, old = curr->next_screen; 
 					((window->scroll_up) && (diff) && 
-					(tmp_ft < window->Num_lines) && 
+					(tmp_ft < last_same) && 
 					(!virtual_lines[tmp_ft]));
 						tmp_ft++)
 				{
-					if ((Comp_line(old, virt) == -1) && (!virtual_lines[tmp_ft]))
+					if ((Comp_line(old, virt) == -1) && (!virtual_lines[from_top]))
 					{
-						for (offset = (tmp_ft - from_top); (offset > 0); offset--)
+						/*
+						 |	Find the bottom of the 
+						 |	area that should be 
+						 |	scrolled.
+						 */
+						for (bottom = tmp_ft, old1 = old, 
+						     new1 = virt, count1 = 0;
+							(bottom < window->Num_lines) && 
+								(Comp_line(old1, new1) <= 0);
+								bottom++, old1 = old1->next_screen, 
+								new1 = new1->next_screen, 
+								count1++)
+							;
+						if (count1 > 3)
 						{
-							old = Delete_line(from_top, window);
-							if (first_same < window->Num_lines)
+							if (String_table[cs__]) /* scrolling region */
 							{
-								old = Insert_line(count1, window);
+								list[1] = from_top;
+								list[0] = min((bottom - 1), (window->Num_lines - 1));
+								String_Out(String_table[cs__], list, 2);
+								Curr_y = Curr_x = -1;
 							}
-							else
+
+							for (offset = (tmp_ft - from_top); (offset > 0); offset--)
 							{
-								Position(window, (window->Num_lines -1), 0);
-								String_Out(String_table[al__], NULL, 0);
+								old = Delete_line(from_top, min((bottom - 1), (window->Num_lines - 1)), window);
+								diff = FALSE;
 							}
-							window->scroll_up--;
-							diff = FALSE;
-						}
-						top_of_win = curscr->first_line;
-						curr = top_of_win;
-						for (offset = 0; offset < from_top; offset++)
-							curr = curr->next_screen;
-						for (offset = from_top, old=curr, new=virt; 
-						   offset < window->Num_lines; 
-						   old=old->next_screen, new=new->next_screen,
-						   offset++)
-						{
-							virtual_lines[offset] = TRUE;
-							if ((similar = Comp_line(old, new)) > 0)
+
+							if (String_table[cs__]) /* scrolling region */
 							{
-								virtual_lines[offset] = FALSE;
+								list[1] = 0;
+								list[0] = LINES;
+								String_Out(String_table[cs__], list, 2);
+								Curr_y = Curr_x = -1;
+							}
+
+							top_of_win = curscr->first_line;
+							curr = top_of_win;
+							for (offset = 0; offset < from_top; offset++)
+								curr = curr->next_screen;
+							for (offset = from_top, old=curr, new=virt; 
+							   offset < window->Num_lines; 
+							   old=old->next_screen, new=new->next_screen,
+							   offset++)
+							{
+								similar = Comp_line(old, new);
+								virtual_lines[offset] = (similar > 0 ? FALSE : TRUE);
 							}
 						}
 					}
@@ -3126,37 +3304,54 @@ doupdate()
 				{
 					if (Comp_line(old, virt) == -1)
 					{
-						for (first_same = window->Num_lines; (first_same > from_top) && (virtual_lines[first_same - 1]); first_same--)
+						/*
+						 |	Find the bottom of the 
+						 |	area that should be 
+						 |	scrolled.
+						 */
+						for (bottom = from_top, old1 = old, 
+						     new1 = virt, count1 = 0;
+							(bottom < window->Num_lines) && 
+								(Comp_line(old1, new1) <= 0);
+								bottom++, old1 = old1->next_screen, 
+								new1 = new1->next_screen, 
+								count1++)
 							;
-						count1 = first_same -  1;
-						for (offset = (from_top - tmp_ft); (offset > 0); offset--)
+						if (count1 > 3)
 						{
-							if (first_same < window->Num_lines)
+							if (String_table[cs__]) /* scrolling region */
 							{
-								old = Delete_line(count1, window);
+								list[1] = tmp_ft;
+								list[0] = min((bottom - 1), (window->Num_lines - 1));
+								String_Out(String_table[cs__], list, 2);
+								Curr_y = Curr_x = -1;
 							}
-							else
+
+							for (offset = (from_top - tmp_ft); (offset > 0); offset--)
 							{
-								Position(window, (window->Num_lines -1), 0);
-								String_Out(String_table[dl__], NULL, 0);
+								old = Insert_line(tmp_ft, min((bottom - 1), (window->Num_lines -1)), window);
+								diff = FALSE;
 							}
-							old = Insert_line(tmp_ft, window);
-							window->scroll_down--;
-							diff = FALSE;
-						}
-						top_of_win = curscr->first_line;
-						curr = top_of_win;
-						for (offset = 0; offset < from_top; offset++)
-							curr = curr->next_screen;
-						for (offset = from_top, old=curr, new=virt; 
-						   offset < window->Num_lines; 
-						   old=old->next_screen, new=new->next_screen,
-						   offset++)
-						{
-							virtual_lines[offset] = TRUE;
-							if ((similar = Comp_line(old, new)) > 0)
+
+							if (String_table[cs__]) /* scrolling region */
 							{
-								virtual_lines[offset] = FALSE;
+								list[1] = 0;
+								list[0] = LINES;
+								String_Out(String_table[cs__], list, 2);
+								Curr_y = Curr_x = -1;
+							}
+
+							top_of_win = curscr->first_line;
+							curr = top_of_win;
+							for (offset = 0; offset < from_top; offset++)
+								curr = curr->next_screen;
+							for (offset = from_top, old=curr, new=virt; 
+							   offset < window->Num_lines; 
+							   old=old->next_screen, new=new->next_screen,
+							   offset++)
+							{
+								similar = Comp_line(old, new);
+								virtual_lines[offset] = (similar > 0 ? FALSE : TRUE);
 							}
 						}
 					}
@@ -3169,6 +3364,7 @@ doupdate()
 			virt = virt->next_screen;
 		}
 	}
+
 	for (from_top = 0, curr = curscr->first_line; from_top < window->SR; from_top++)
 		curr = curr->next_screen;
 	top_of_win = curr;
@@ -3289,6 +3485,7 @@ doupdate()
 	Position(window, window->LY, window->LX);
 }
 
+void 
 Position(window, row, col)	/* position the cursor for output on the screen	*/
 WINDOW *window;
 int row;
@@ -3318,6 +3515,7 @@ int col;
 	}
 }
 
+void 
 Char_del(line, attrib, offset, maxlen)	/* delete chars from line	*/
 char *line;
 char *attrib;
@@ -3334,6 +3532,7 @@ int maxlen;
 	String_Out(String_table[dc__], NULL, 0);
 }
 
+void 
 Char_ins(line, attrib, newc, newatt, offset, maxlen)	/* insert chars in line	*/
 char *line;
 char *attrib;
@@ -3357,6 +3556,7 @@ int maxlen;
 	Char_out(newc, newatt, line, attrib, offset);
 }
 
+void 
 attribute_on()
 {
 	if (String_table[sa__])
@@ -3368,6 +3568,7 @@ attribute_on()
 		String_Out(String_table[so__], NULL, 0);
 }
 
+void 
 attribute_off()
 {
 	if (String_table[me__])
@@ -3381,6 +3582,7 @@ attribute_off()
 		String_Out(String_table[se__], NULL, 0);
 }
 
+void 
 Char_out(newc, newatt, line, attrib, offset)	/* output character with proper attribute	*/
 char newc;
 char newatt;
